// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Xml;

namespace PiDuce.Common
{
	public enum BasicTypeKind { BOOL, INT, FLOAT, STRING }

	public interface IBasicType
	{
		BasicTypeKind Kind { get; }
		string ToString();
		void ToXml(XmlWriter writer);
	}

	public abstract class AbstractBasicType : IBasicType
	{
		private readonly BasicTypeKind kind;

		public BasicTypeKind Kind { get { return kind; } }

		public AbstractBasicType(BasicTypeKind kind)
		{ this.kind = kind; }

		public override bool Equals(Object obj)
		{ return (obj is AbstractBasicType) && (kind == ((AbstractBasicType) obj).kind); }

		public override int GetHashCode()
		{ return (int) kind; }

		public abstract void ToXml(XmlWriter writer);
	}

	public class BoolType : AbstractBasicType
	{
		public BoolType() : base(BasicTypeKind.BOOL)
		{ }

		public override string ToString()
		{ return "bool"; }

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("bool");
			writer.WriteEndElement(); // </bool>
		}
	}

	public class IntType : AbstractBasicType
	{
		public IntType() : base(BasicTypeKind.INT)
		{ }

		public override string ToString()
		{ return "int"; }

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("int");
			writer.WriteEndElement(); // </int>
		}
	}

	public class FloatType : AbstractBasicType
	{
		public FloatType() : base(BasicTypeKind.FLOAT)
		{ }

		public override string ToString()
		{ return "float"; }

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("float");
			writer.WriteEndElement(); // </float>
		}
	}

	public class StringType : AbstractBasicType
	{
		public StringType() : base(BasicTypeKind.STRING)
		{ }

		public override string ToString()
		{ return "string"; }

		public override void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("string");
			writer.WriteEndElement(); // </string>
		}
	}

	public interface IBasicTypeLiteral
	{
		string ToString();
		void ToXml(XmlWriter writer);
		IBasicType TypeOf();
		int GetValueHashCode();
	}

	public abstract class BasicTypeLiteral<T> : IBasicTypeLiteral
	{
		private readonly T value;

		public T Value { get { return value; } }

		public BasicTypeLiteral(T value)
		{ this.value = value; }

		public override string ToString()
		{ return value.ToString(); }

		public void ToXml(XmlWriter writer)
		{
			writer.WriteStartElement("literal");
			writer.WriteAttributeString("type", TypeOf().ToString());
			writer.WriteString(this.ToString());
			writer.WriteEndElement();
		}

		public override bool Equals(Object obj)
		{
			if (obj is BasicTypeLiteral<T>)
				return value.Equals(((BasicTypeLiteral<T>) obj).value);
			else
				return false;
		}

    public override int GetHashCode()
    {
      return base.GetHashCode();
    }

		public int GetValueHashCode()
		{ return value.GetHashCode(); }

		public abstract IBasicType TypeOf();
	}

	public class BoolLiteral : BasicTypeLiteral<bool>
	{
		public BoolLiteral(bool value) : base(value)
		{ }

		public override IBasicType TypeOf()
		{ return new BoolType(); }
	}

	public class IntLiteral : BasicTypeLiteral<int>
	{
		public IntLiteral(int value) : base(value)
		{ }

		public override IBasicType TypeOf()
		{ return new IntType(); }
	}
	
	public class FloatLiteral : BasicTypeLiteral<float>
	{
		public FloatLiteral(float value) : base(value)
		{ }

		public override IBasicType TypeOf()
		{ return new FloatType(); }
	}
	
	public class StringLiteral : BasicTypeLiteral<string>
	{
		public StringLiteral(string value) : base(value)
		{ }

		public override IBasicType TypeOf()
		{ return new StringType(); }
	}
}

