// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Xml;
using System.Collections;
using System.Diagnostics;
using PiDuce.Common;
using PiDuce.Types;
using System.Collections.Generic;

using IPattern = PiDuce.Common.ISchema;

namespace PiDuce.Compiler
{
	public class AllocateContext
	{
		private readonly Stack size;
		private int currentSize; // size of the environment 

		public AllocateContext()
		{
			this.size = new Stack();
			this.currentSize = 0;
		}

		public void Enter()
		{ size.Push(currentSize); }

		public void Exit()
		{ currentSize = (int) size.Pop(); }

		public int NextIndex { get { return currentSize++; } }
		public int Size { get { return currentSize; } }

		public void Alloc(int n)
		{
			Debug.Assert(n >= currentSize);
			currentSize = n;
		}
	}

	// EXPRESSIONS
	public class CompileExpressionVisitor : IExpressionVisitor
	{
		private readonly ExpressionSchemaContext schemaContext;
		private readonly XmlWriter code;

		public CompileExpressionVisitor(ExpressionSchemaContext schemaContext, XmlWriter code)
		{ 
			this.code = code;
			this.schemaContext = schemaContext;
		}

		public void Compile(IExpression expr)
		{ expr.Accept(this); }

		public void VisitVoidExpr(VoidExpr expr)
		{
      code.WriteStartElement("void");
			code.WriteEndElement(); // </void>
		}

		public void VisitLiteralExpr(LiteralExpr expr)
		{
			code.WriteStartElement("literal");
			code.WriteAttributeString("type", expr.Content.TypeOf().ToString());
			code.WriteString(expr.Content.ToString());
			code.WriteEndElement(); // </literal>
		}

		public void VisitVariableExpr(VariableExpr expr)
		{
			code.WriteStartElement("load");
			code.WriteAttributeString("index", ((ValueEntry) expr.Entry).Index.ToString());
      code.WriteAttributeString("name", expr.Entry.Name);
			code.WriteEndElement(); // </load>
		}

		public void VisitSelectExpr(SelectExpr expr)
		{
			code.WriteStartElement("select");
			code.WriteAttributeString("field", expr.Field);
			expr.Content.Accept(this);
			code.WriteEndElement(); // </select>
		}

		public void VisitRecordExpr(RecordExpr expr)
		{
			code.WriteStartElement("record");
			foreach (KeyValuePair<string, IExpression> pair in expr.Fields) {
				code.WriteStartElement("field");
				code.WriteAttributeString("name", pair.Key);
				pair.Value.Accept(this);
				code.WriteEndElement(); // </field>
			}
			code.WriteEndElement(); // </record>
		}

		public void VisitLabelledExpr(LabelledExpr expr)
		{
			code.WriteStartElement("labelled");
			code.WriteAttributeString("label", expr.Label);
			code.WriteStartElement("attributes");
			expr.Attributes.Accept(this);
			code.WriteEndElement(); // </attributes>
			code.WriteStartElement("content");
			expr.Content.Accept(this);
			code.WriteEndElement(); // </content>
			code.WriteEndElement(); // </labelled>
		}

		public void VisitSequenceExpr(SequenceExpr expr)
		{
			code.WriteStartElement("seq");
			Compile(expr.Head);
			Compile(expr.Tail);
			code.WriteEndElement(); // </seq>
		}

		public void VisitBinaryOpExpr(BinaryOpExpr expr)
		{
			code.WriteStartElement("binop");
			switch (expr.Op) {
				case BinaryOp.PLUS: code.WriteAttributeString("op", "add"); break;
        case BinaryOp.MINUS: code.WriteAttributeString("op", "sub"); break;
				case BinaryOp.TIMES: code.WriteAttributeString("op", "mul"); break;
				case BinaryOp.DIVIDE: code.WriteAttributeString("op", "div"); break;
				case BinaryOp.EQ: code.WriteAttributeString("op", "eq"); break;
				case BinaryOp.NE: code.WriteAttributeString("op", "neq"); break;
				case BinaryOp.GT: code.WriteAttributeString("op", "gt"); break;
				case BinaryOp.GE: code.WriteAttributeString("op", "ge"); break;
				case BinaryOp.LT: code.WriteAttributeString("op", "lt"); break;
				case BinaryOp.LE: code.WriteAttributeString("op", "le"); break;
				case BinaryOp.LAND: code.WriteAttributeString("op", "land"); break;
				case BinaryOp.LOR: code.WriteAttributeString("op", "lor"); break;
				default: Debug.Assert(false); break;
			}
			Compile(expr.Left);
			Compile(expr.Right);
			code.WriteEndElement(); // </binop>
		}

		public void VisitUnaryOpExpr(UnaryOpExpr expr)
		{
			code.WriteStartElement("unop");
			switch (expr.Op) {
				case UnaryOp.UMINUS: code.WriteAttributeString("op", "minus"); break;
				case UnaryOp.LNOT: code.WriteAttributeString("op", "lnot"); break;
				default: Debug.Assert(false); break;
			}
			Compile(expr.Content);
			code.WriteEndElement(); // </unop>
		}
	}

	// PATTERNS
	public class AllocateNamesPatternVisitor : AbstractSchemaVisitor
	{
		private readonly AllocateContext context;

		public AllocateNamesPatternVisitor(AllocateContext context)
		{ this.context = context; }

		public void AllocateNames(IPattern pattern)
		{ pattern.Accept(this); }

		public override void VisitBindSchema(BindSchema s)
		{
			Debug.Assert(s.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) s.Entry;
			if (entry.Index < 0 && entry.Used)
				entry.Index = context.NextIndex;
			base.VisitBindSchema(s);
		}
	}

	public class CompilePatternVisitor : SchemaToXmlVisitor
	{
		public CompilePatternVisitor(XmlWriter writer) : base(writer)
		{ }

		public void Compile(IPattern pattern)
		{ pattern.Accept(this); }

		public override void VisitBindSchema(BindSchema s)
		{
			Debug.Assert(s.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) s.Entry;
			if (entry.Used) {
				writer.WriteStartElement("bind");
				writer.WriteAttributeString("index", entry.Index.ToString());
				writer.WriteAttributeString("name", entry.Name);
			}
			s.Content.Accept(this);
			if (entry.Used)
				writer.WriteEndElement(); // </bind>
		}
	}

	// PROCESSES
	public class AllocateNamesProcessVisitor : IProcessVisitor
	{
		private readonly AllocateContext context;
		private readonly AllocateNamesPatternVisitor patternVisitor;

		public AllocateNamesProcessVisitor(AllocateContext context)
		{
			this.context = context;
			this.patternVisitor = new AllocateNamesPatternVisitor(context);
		}

		public void AllocateNames(IProcess p)
		{ p.Accept(this); }

		private void AllocateNames(IPattern p)
		{ patternVisitor.AllocateNames(p); }

		public void VisitNilProcess(NilProcess p)
		{ }

		public void VisitSpawnProcess(SpawnProcess p)
		{
			int parentEnvSize = context.Size;
			context.Enter();
			AllocateNames(p.Left);
			int leftEnvSize = context.Size;
			p.SetEnvSize(parentEnvSize, leftEnvSize);
			context.Exit();
			AllocateNames(p.Right);
		}

		public void VisitOutputProcess(OutputProcess p)
		{ }

		public void VisitInputProcess(InputProcess p)
		{
			AllocateNames(p.Pattern);
			AllocateNames(p.Continuation);
		}

		public void VisitJoinProcess(JoinProcess p)
		{
			foreach (JoinPatternRow row in p.Rows) {
				foreach (JoinPatternAtom atom in row.Atoms)
					AllocateNames(atom.Pattern);
				AllocateNames(row.Continuation);
			}
		}

		public void VisitMatchProcess(MatchProcess p)
		{
			int maxSize = 0;
			foreach (MatchProcess.Branch branch in p.Branches) {
				context.Enter();
				AllocateNames(branch.Pattern);
				AllocateNames(branch.Process);
				int size = context.Size;
				context.Exit();
				maxSize = System.Math.Max(maxSize, size);
			}
			context.Alloc(maxSize);
		}

		public void VisitImportProcess(ImportProcess p)
		{
			Debug.Assert(p.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) p.Entry;
			if (entry.Used)
				entry.Index = context.NextIndex;
			AllocateNames(p.Continuation);
		}

		public void VisitNewProcess(NewProcess p)
		{
			Debug.Assert(p.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) p.Entry;
			if (entry.Used)
				entry.Index = context.NextIndex;
			AllocateNames(p.Continuation);
		}

		public void VisitSchemaProcess(SchemaProcess p)
		{ AllocateNames(p.Continuation); }
	}

	public class CompileProcessVisitor : IProcessVisitor
	{
		private readonly XmlWriter code;
		private readonly SchemaToXmlVisitor schemaVisitor;
		private readonly CompilePatternVisitor patternVisitor;
		private readonly CompileExpressionVisitor exprVisitor;

		public CompileProcessVisitor(ExpressionSchemaContext schemaContext, XmlWriter code)
		{
			this.code = code;
			this.schemaVisitor = new SchemaToXmlVisitor(code);
			this.patternVisitor = new CompilePatternVisitor(code);
			this.exprVisitor = new CompileExpressionVisitor(schemaContext, code);
		}

		public void Compile(IProcess p)
		{ p.Accept(this); }

		private void Compile(IExpression expr)
		{ exprVisitor.Compile(expr); }

		private void CompileSchema(ISchema s)
		{ schemaVisitor.ToXml(s); }

		private void CompilePattern(IPattern pattern)
		{ patternVisitor.Compile(pattern); }

		public void VisitNilProcess(NilProcess p)
		{ }

		public void VisitSpawnProcess(SpawnProcess p)
		{
			code.WriteStartElement("spawn");
			code.WriteStartElement("process");
			code.WriteAttributeString("parent-env-size", p.ParentEnvSize.ToString());
			code.WriteAttributeString("local-env-size", p.LocalEnvSize.ToString());
			Compile(p.Left);
			code.WriteEndElement(); // </process>
			Compile(p.Right);
      code.WriteEndElement(); // </spawn>
		}

		public void VisitOutputProcess(OutputProcess p)
		{	
			code.WriteStartElement("send");
			code.WriteStartElement("destination");
			Compile(p.Destination);
			code.WriteEndElement(); // </destination>
			code.WriteStartElement("expression");
			Compile(p.Expression);
			code.WriteEndElement(); // </expression>
			code.WriteEndElement(); // </send>
		}

		public void VisitInputProcess(InputProcess p)
		{			
			if (p.Service)
				code.WriteStartElement("service");
			else
        code.WriteStartElement("receive");
			code.WriteStartElement("source");
			Compile(p.Source);
			code.WriteEndElement(); // </source>
			code.WriteStartElement("pattern");
			CompilePattern(p.Pattern);
			code.WriteEndElement(); // </pattern>
			Compile(p.Continuation);
			code.WriteEndElement(); // </service> or </receive>
		}

		public void VisitJoinProcess(JoinProcess p)
		{
			code.WriteStartElement("join-select");
			code.WriteAttributeString("service", p.Service.ToString().ToLower());

			int nextChannelIndex = 0;
			IDictionary<string, int> chEntries = new Dictionary<string, int>();
			foreach (JoinPatternRow row in p.Rows) {
				code.WriteStartElement("branch");
				foreach (JoinPatternAtom atom in row.Atoms) {
					code.WriteStartElement("receive");
          if (!chEntries.ContainsKey(atom.Entry.Name))
          {
            atom.Index = nextChannelIndex++;
            chEntries[atom.Entry.Name] = atom.Index;
          }
          else atom.Index = chEntries[atom.Entry.Name];
          Debug.Assert(atom.Index >= 0);
					code.WriteAttributeString("index", atom.Index.ToString());
          code.WriteStartElement("source");
          Compile(atom.Source);
          code.WriteEndElement();
					code.WriteStartElement("pattern");
					CompilePattern(atom.Pattern);
					code.WriteEndElement(); // </pattern>
					code.WriteEndElement(); // </receive>
				}
				code.WriteStartElement("process");
				Compile(row.Continuation);
				code.WriteEndElement(); // </process>
				code.WriteEndElement(); // </branch>
			}

			code.WriteEndElement(); // </join-select>
		}

		public void VisitMatchProcess(MatchProcess p)
		{
			code.WriteStartElement("match");			
			code.WriteStartElement("expression");
			Compile(p.Expression);
			code.WriteEndElement(); // </expression>
			foreach (MatchProcess.Branch branch in p.Branches) {
				code.WriteStartElement("branch");				
				code.WriteStartElement("pattern");
				CompilePattern(branch.Pattern);
				code.WriteEndElement(); // </pattern>
				if (branch.Condition != null) {
					code.WriteStartElement("when");
					Compile(branch.Condition);
					code.WriteEndElement(); // </when>
				}
				code.WriteStartElement("process");
				Compile(branch.Process);
				code.WriteEndElement(); // </process>
				code.WriteEndElement(); // </branch>
			}
			code.WriteEndElement(); // </match>
		}

		public void VisitImportProcess(ImportProcess p)
		{
			Debug.Assert(p.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) p.Entry;
      if (entry.Used)
      {
        code.WriteStartElement("import");
        code.WriteAttributeString("target", entry.Index.ToString());
        code.WriteAttributeString("wsdl", p.Wsdl);
				code.WriteStartElement("operations");
				Debug.Assert(p.Entry.Schema is ServiceSchema);
        foreach (KeyValuePair<string, ISchema> opEntry in ((ServiceSchema) p.Entry.Schema).Operations)
        {
          code.WriteStartElement("operation");
          code.WriteAttributeString("name", opEntry.Key);
          CompileSchema(opEntry.Value);
          code.WriteEndElement(); // </operation>
        }
				code.WriteEndElement(); // </operation>
      }

			Compile(p.Continuation);

			if (entry.Used)
				code.WriteEndElement(); // </import>
		}

		public void VisitNewProcess(NewProcess p)
		{
			Debug.Assert(p.Entry is ValueEntry);
			ValueEntry entry = (ValueEntry) p.Entry;
      if (entry.Used)
      {
        code.WriteStartElement("new");
        code.WriteAttributeString("target", entry.Index.ToString());
        code.WriteAttributeString("target-name", entry.Name.ToString());
        if (p.ChannelLocation != "")
          code.WriteAttributeString("location", p.ChannelLocation);
				code.WriteStartElement("operations");
				Debug.Assert(p.Entry.Schema is ServiceSchema);
        foreach (KeyValuePair<string, ISchema> operation in p.Operations)
        {
          code.WriteStartElement("operation");
          code.WriteAttributeString("name", operation.Key);
          if (p.ExportedCapabilites[operation.Key] == ChannelType.CAPABILITY.IN)
					  code.WriteAttributeString("exported-capability", "I");
          else if (p.ExportedCapabilites[operation.Key] == ChannelType.CAPABILITY.OUT)
            code.WriteAttributeString("exported-capability", "O");
          else if (p.ExportedCapabilites[operation.Key] == ChannelType.CAPABILITY.INOUT)
            code.WriteAttributeString("exported-capability", "IO");
          CompileSchema(operation.Value);
          code.WriteEndElement(); // </operation>
        }
				code.WriteEndElement(); // </operations>
      }
			Compile(p.Continuation);

			if (entry.Used)
				code.WriteEndElement(); // </new>
		}

		public void VisitSchemaProcess(SchemaProcess p)
		{
			foreach (SchemaProcess.Declaration decl in p.Declarations) {
#if false
				// TODO: all named schemas are exported in the appropriate section
				// but in this way the structure of the original process is not
				// preserved. Named schemas should be exported separately
				// (and be given fresh names) and then a schemadec process should
				// just make a reference to that fresh name
				if (decl.Entry.Used) {
					dec.WriteStartElement("schema");
					if (decl.Entry.Name != decl.Entry.FreshName)
						dec.WriteAttributeString("original-name", decl.Entry.Name);
					dec.WriteAttributeString("name", decl.Entry.FreshName);
					Compile(decl.Schema);
					dec.WriteEndElement(); // </schema>
				}
#endif
			}

			Compile(p.Continuation);
		}		 
	}
}
