// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Text.RegularExpressions;
using System.Collections;

namespace PiDuce.Compiler
{
	public class GenericValue
	{
		private readonly int location;

		public int Location { get { return location; } }

		public GenericValue(int location)
		{ this.location = location; }
	}

	public class IntValue : GenericValue
	{
		private readonly int val;

		public int Value { get { return val; } }

		public IntValue(int val, int location) : base(location)
		{ this.val = val; }
	}

	public class FloatValue : GenericValue
	{
		private readonly float val;

		public float Value { get { return val; } }

		public FloatValue(float val, int location) : base(location)
		{ this.val = val; }
	}

	public class StringValue : GenericValue
	{
		private readonly String val;

		public String Value { get { return val; } }

		public StringValue(String val, int location) : base(location)
		{ this.val = val; }
	}
    
	public interface IToken
	{		
		/* integer returned by Token's jay class
		 * that correspond at this keyword*/
		int Token { get; }
		/* return true if <code>str</code>
		 * match with this token */
		bool Match(string str, ref int length);
	}

	internal class SimpleToken : IToken	
	{	
		private readonly int token;
		private readonly String val;

		public int Token { get { return token; } }

		public SimpleToken(string val, int token)
		{
			this.val = val;
			this.token = token;
		}

		public bool Match(String str, ref int length)
		{
			length = val.Length;
			return (str.StartsWith(val));
		}
	}

	internal class RegexpToken : IToken	
	{	
		private readonly int token;
		private readonly Regex val;

		public int Token { get { return token; } }

		public RegexpToken(string pattern, int token)
		{
			this.val = new Regex("^(" + pattern + "){1}");
			this.token = token;
		}

		public bool Match(String str, ref int length)
		{
			Match m = val.Match(str);
			length = m.Length;

			return m.Success;
		}
	}
	
	public class Lexer : yyParser.yyInput
	{
		public Object Value;
		private static ArrayList tokens;
    public static ArrayList Tokens { get { return tokens; } }
		private static Hashtable keywords;
    public static Hashtable Keywords { get { return keywords; } }
		private static Regex annotation = new Regex("^(//.*\n)");
		private string m_input = "";
		//line number
		public int location;

    static Lexer() 
    {
      tokens = new ArrayList();
      keywords = new Hashtable();

      tokens.Add(new RegexpToken("\"([^\"]*|(\\\")*)*\"", Token.TEXT));
      tokens.Add(new RegexpToken("([0-9]+\\.[0-9]*)|([0-9]*\\.[0-9]+)", Token.FLOAT));
      tokens.Add(new RegexpToken("[0-9][0-9]*", Token.INT));
      tokens.Add(new RegexpToken("[A-Za-z][A-Za-z0-9_]*", Token.ID));
      tokens.Add(new RegexpToken("`[A-Za-z_]['\\-''.'A-Za-z0-9_]*", Token.ID));
      tokens.Add(new SimpleToken("->", Token.ARROW));
      tokens.Add(new SimpleToken("=>", Token.DARROW));
      tokens.Add(new SimpleToken("!", Token.EMARK));
      tokens.Add(new SimpleToken("?", Token.QMARK));
      tokens.Add(new SimpleToken("&", Token.AMPERSAND));
      tokens.Add(new SimpleToken("-", Token.MINUS));
      tokens.Add(new SimpleToken("==", Token.EQEQ));
      tokens.Add(new SimpleToken("=", Token.EQ));
      tokens.Add(new SimpleToken("^IO", Token.INOUT_CAP));
      tokens.Add(new SimpleToken("^I", Token.IN_CAP));
      tokens.Add(new SimpleToken("^O", Token.OUT_CAP));
      tokens.Add(new SimpleToken("!=", Token.NE));
      tokens.Add(new SimpleToken("{", Token.LBRACE));
      tokens.Add(new SimpleToken("}", Token.RBRACE));
      tokens.Add(new SimpleToken("(", Token.LPAREN));
      tokens.Add(new SimpleToken(")", Token.RPAREN));
      tokens.Add(new SimpleToken("[", Token.LBRACK));
      tokens.Add(new SimpleToken("]", Token.RBRACK));
      tokens.Add(new SimpleToken(":", Token.COLON));
      tokens.Add(new SimpleToken(";", Token.SEMICOLON));
      tokens.Add(new SimpleToken(".", Token.DOT));
      tokens.Add(new SimpleToken("+", Token.PLUS));
      tokens.Add(new SimpleToken("*", Token.TIMES));
      tokens.Add(new SimpleToken("/", Token.DIVIDE));
      tokens.Add(new SimpleToken("\\", Token.BACKSLASH));
      tokens.Add(new SimpleToken("=", Token.EQ));
      tokens.Add(new SimpleToken(">=", Token.GE));
      tokens.Add(new SimpleToken("<=", Token.LE));
      tokens.Add(new SimpleToken(">", Token.GT));
      tokens.Add(new SimpleToken("<", Token.LT));
      tokens.Add(new SimpleToken(",", Token.COMMA));
      tokens.Add(new SimpleToken("|", Token.OR));
      tokens.Add(new SimpleToken("||", Token.LOR));
      tokens.Add(new SimpleToken("&&", Token.LAND));
      tokens.Add(new SimpleToken("~", Token.TILDE));
      //tokens.Add(new SimpleToken("@", Token.AT));
      
      keywords.Add("wsdl", Token.KW_WSDL);
      keywords.Add("and", Token.KW_AND);
      keywords.Add("location", Token.KW_LOCATION);
      keywords.Add("spawn", Token.KW_SPAWN);
      keywords.Add("match", Token.KW_MATCH);
      keywords.Add("when", Token.KW_WHEN);
      keywords.Add("join", Token.KW_JOIN);
      keywords.Add("bool", Token.KW_BOOL);
      keywords.Add("int", Token.KW_INT);
      keywords.Add("float", Token.KW_FLOAT);
      keywords.Add("string", Token.KW_STRING);
      keywords.Add("let", Token.KW_LET);
      keywords.Add("new", Token.KW_NEW);
      keywords.Add("in", Token.KW_IN);
      keywords.Add("with", Token.KW_WITH);
      keywords.Add("import", Token.KW_IMPORT);
      keywords.Add("schema", Token.KW_SCHEMA);
      keywords.Add("nil", Token.KW_NIL);
      keywords.Add("false", Token.KW_FALSE);
      keywords.Add("true", Token.KW_TRUE);
      keywords.Add("if", Token.KW_IF);
      keywords.Add("then", Token.KW_THEN);
      keywords.Add("else", Token.KW_ELSE);
    }

		public Lexer(string m_input)
		{	
			location = 1;
			this.m_input = m_input;
		}
		
		public Object value() {
			return Value;
		}

		public bool advance()
		{
			while (m_input.Length != 0) {
				while (m_input.Length != 0 && (m_input[0] == '\n' || m_input[0] == '\t' || m_input[0] == ' ')) {
					if (m_input[0] == '\n') location++;
					m_input = m_input.Remove(0, 1);
				}
				Match m = annotation.Match(m_input);
				if (m.Success) {
					m_input = m_input.Remove(0, m.Length);
					location++;				
				} else
					break;
			}

			return (m_input.Length != 0);
		}

		public int token()
		{
			foreach (IToken tmp in tokens) {
				int length = 0;
				if (tmp.Match(m_input, ref length)) {
				//Console.WriteLine("matched token {0} {1}", tmp.Token, m_input.Substring(0, length));
					string lexeme = m_input.Substring(0, length);
					switch (tmp.Token) {
						case Token.TEXT:
							lexeme = lexeme.Substring(1, length - 2);
							Value = new StringValue(lexeme.Replace("\\\"","\""), location);
							break;
						case Token.INT:
							Value = new IntValue(Int32.Parse(lexeme), location);
							break;
						case Token.FLOAT:
							Value = new FloatValue(Single.Parse(lexeme), location);
							break;
						case Token.ID:
							if (keywords.Contains(lexeme)) {
								Value = new GenericValue(location);
								m_input = m_input.Remove(0, length);
								return (int) keywords[lexeme];
							}

							if (!m_input.StartsWith("`"))
								Value = new StringValue(lexeme, location);
							else
								Value = new StringValue(lexeme.Substring(1, length - 1), location);
							break;
						default:
							Value = new GenericValue(location);
							break;
					}

					m_input = m_input.Remove(0, length);
					return tmp.Token;
				}
			}

			return Token.ERROR;
		}
	}
}
