// created by jay 0.7 (c) 1998 Axel.Schreiner@informatik.uni-osnabrueck.de

#line 2 "Parser.jay"
// -*-csharp-*-
// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.IO;
using System.Text;
using System.Collections.Generic;
using System.Diagnostics;
using PiDuce.Types;
using PiDuce.Common;

// the following aliases are needed because currently Jay does not recognize
// generic types within < ... >

using IPattern = PiDuce.Common.ISchema;
using ISet_string = PiDuce.Common.ISet<string>;
using IList_IProcess = System.Collections.Generic.IList<PiDuce.Common.IProcess>;
using IList_JoinPatternRow = System.Collections.Generic.IList<PiDuce.Common.JoinPatternRow>;
using IList_JoinPatternAtom = System.Collections.Generic.IList<PiDuce.Common.JoinPatternAtom>;
using IDictionary_FieldDecList = System.Collections.Generic.IDictionary<string, PiDuce.Common.ISchema>;
using IDictionary_FieldDefList = System.Collections.Generic.IDictionary<string, PiDuce.Common.IExpression>;
using IDictionary_IExpression = System.Collections.Generic.IDictionary<string, PiDuce.Common.IExpression>;
using IDictionary_ISchema = System.Collections.Generic.IDictionary<string, PiDuce.Common.ISchema>;
using IList_SchemaProcess_Declaration = System.Collections.Generic.IList<PiDuce.Common.SchemaProcess.Declaration>;
using IList_MatchProcess_Branch = System.Collections.Generic.IList<PiDuce.Common.MatchProcess.Branch>;

namespace PiDuce.Compiler {

	public class SyntaxError : Exception
	{
		public SyntaxError(string msg) : base(msg)
		{ }
	}

	public class LocationContext
	{
		private readonly Stack<int> locations;

		public LocationContext()
		{
			this.locations = new Stack<int>();
		}

		public int GetLocation(Object obj)
		{
			if (PiDuceParser.HasLocation(obj))
				return PiDuceParser.GetLocation(obj);
			else if (locations.Count > 0)
				return locations.Peek();
			else
				return -1;
		}

		public void Enter(Object obj)
		{
			if (PiDuceParser.HasLocation(obj))
				locations.Push(PiDuceParser.GetLocation(obj));
		}

		public void Exit(Object obj)
		{
			if (PiDuceParser.HasLocation(obj))
				locations.Pop();
		}
	}

	public class PiDuceParser {
  	private Lexer lexer;
  	private static IDictionary<Object, int> locationMap;
  	public static int yacc_verbose_flag = 2;

#line default

  /** error output stream.
      It should be changeable.
    */
  public System.IO.TextWriter ErrorOutput = System.Console.Out;

  /** simplified error message.
      @see <a href="#yyerror(java.lang.String, java.lang.String[])">yyerror</a>
    */
  public void yyerror (string message) {
    yyerror(message, null);
  }

  /** (syntax) error message.
      Can be overwritten to control message format.
      @param message text to be displayed.
      @param expected vector of acceptable tokens, if available.
    */
  public void yyerror (string message, string[] expected) {
    CompilerOutput.Error("syntax error", lexer.location);
    if ((yacc_verbose_flag > 0) && (expected != null) && (expected.Length  > 0)) {
      ErrorOutput.Write (message+", expecting");
      for (int n = 0; n < expected.Length; ++ n)
        ErrorOutput.Write (" "+expected[n]);
        ErrorOutput.WriteLine ();
    } else
      ErrorOutput.WriteLine (message);
  }

  /** debugging support, requires the package jay.yydebug.
      Set to null to suppress debugging messages.
    */
//t  internal yydebug.yyDebug debug;

  protected static  int yyFinal = 20;
//t  public static  string [] yyRule = {
//t    "$accept : process",
//t    "process : KW_NIL",
//t    "process : expr EMARK expr",
//t    "process : expr QMARK star_opt pattern DARROW process",
//t    "process : KW_SCHEMA schema_decl_ne_list KW_IN process",
//t    "process : KW_LET pattern EQ expr KW_IN process",
//t    "process : KW_NEW ID operations location_opt KW_IN process",
//t    "process : KW_IMPORT ID operations location_opt KW_IN process",
//t    "process : KW_SPAWN LBRACE process RBRACE process",
//t    "process : KW_MATCH expr KW_WITH LBRACE branch_ne_list RBRACE",
//t    "process : star_opt KW_JOIN LBRACE join_pattern_row_ne_list RBRACE",
//t    "process : KW_IF expr KW_THEN process else_opt",
//t    "process : LBRACE process LBRACE",
//t    "else_opt :",
//t    "else_opt : KW_ELSE process",
//t    "operations : LBRACE field_dec_ne_list RBRACE",
//t    "operations : COLON schema",
//t    "field_dec_list :",
//t    "field_dec_list : field_dec_ne_list",
//t    "field_dec_ne_list : ID COLON schema",
//t    "field_dec_ne_list : ID COLON schema SEMICOLON field_dec_ne_list",
//t    "star_opt :",
//t    "star_opt : TIMES",
//t    "join_pattern_row_ne_list : join_pattern_row",
//t    "join_pattern_row_ne_list : join_pattern_row_ne_list OR join_pattern_row",
//t    "join_pattern_row : join_pattern_atom_ne_list DARROW process",
//t    "join_pattern_atom_ne_list : join_pattern_atom",
//t    "join_pattern_atom_ne_list : join_pattern_atom_ne_list AMPERSAND join_pattern_atom",
//t    "join_pattern_atom : expr QMARK pattern_opt",
//t    "schema_decl_ne_list : schema_decl",
//t    "schema_decl_ne_list : schema_decl_ne_list KW_AND schema_decl",
//t    "schema_decl : ID EQ schema",
//t    "branch_ne_list : branch",
//t    "branch_ne_list : branch_ne_list OR branch",
//t    "branch : pattern when_opt DARROW process",
//t    "when_opt :",
//t    "when_opt : KW_WHEN expr",
//t    "pattern_opt :",
//t    "pattern_opt : pattern",
//t    "expr_opt :",
//t    "expr_opt : expr",
//t    "literal : INT",
//t    "literal : FLOAT",
//t    "literal : TEXT",
//t    "literal : KW_FALSE",
//t    "literal : KW_TRUE",
//t    "field_def_list :",
//t    "field_def_list : field_def_ne_list",
//t    "field_def_ne_list : ID EQ expr",
//t    "field_def_ne_list : ID EQ expr SEMICOLON field_def_ne_list",
//t    "record_opt :",
//t    "record_opt : LBRACE field_def_list RBRACE",
//t    "expr : LPAREN expr_opt RPAREN",
//t    "expr : ID",
//t    "expr : expr DOT ID",
//t    "expr : LBRACE field_def_list RBRACE",
//t    "expr : literal",
//t    "expr : ID record_opt LBRACK expr_opt RBRACK",
//t    "expr : expr PLUS expr",
//t    "expr : expr MINUS expr",
//t    "expr : expr TIMES expr",
//t    "expr : expr DIVIDE expr",
//t    "expr : expr LAND expr",
//t    "expr : expr LOR expr",
//t    "expr : expr EQEQ expr",
//t    "expr : expr NE expr",
//t    "expr : expr GT expr",
//t    "expr : expr GE expr",
//t    "expr : expr LT expr",
//t    "expr : expr LE expr",
//t    "expr : expr COMMA expr",
//t    "expr : EMARK expr",
//t    "expr : MINUS expr",
//t    "label_set : ID",
//t    "label_set : TILDE",
//t    "label_set : LBRACE label_comma_list RBRACE",
//t    "label_set : LBRACE TILDE BACKSLASH label_ne_comma_list RBRACE",
//t    "label_comma_list :",
//t    "label_comma_list : label_ne_comma_list",
//t    "label_ne_comma_list : ID",
//t    "label_ne_comma_list : label_ne_comma_list COMMA ID",
//t    "schema_opt :",
//t    "schema_opt : schema",
//t    "basic_type : KW_BOOL",
//t    "basic_type : KW_INT",
//t    "basic_type : KW_FLOAT",
//t    "basic_type : KW_STRING",
//t    "record_schema_opt :",
//t    "record_schema_opt : LBRACE field_dec_list RBRACE",
//t    "schema : basic_type",
//t    "schema : literal",
//t    "schema : LT schema_opt GT capability_opt",
//t    "schema : schema ARROW capability_opt schema",
//t    "schema : schema OR schema",
//t    "schema : ID",
//t    "schema : ID COLON schema",
//t    "schema : LPAREN RPAREN",
//t    "schema : LPAREN schema RPAREN",
//t    "schema : label_set record_schema_opt LBRACK schema_opt RBRACK",
//t    "schema : LBRACE field_dec_list RBRACE",
//t    "schema : schema COMMA schema",
//t    "schema : schema TIMES",
//t    "schema : schema PLUS",
//t    "schema : schema LBRACE INT RBRACE",
//t    "schema : schema LBRACE INT COMMA INT RBRACE",
//t    "schema : schema QMARK",
//t    "capability : IN_CAP",
//t    "capability : OUT_CAP",
//t    "capability : INOUT_CAP",
//t    "capability_opt :",
//t    "capability_opt : capability",
//t    "location : KW_LOCATION EQ TEXT",
//t    "location_opt :",
//t    "location_opt : location",
//t    "pattern : schema",
//t  };
  protected static  string [] yyNames = {    
    "end-of-file",null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,"ERROR","TEXT","INT","FLOAT","ID",
    "ARROW","DARROW","LBRACE","RBRACE","LPAREN","RPAREN","LBRACK",
    "RBRACK","QMARK","EMARK","COLON","SEMICOLON","DOT","PLUS","MINUS",
    "TIMES","DIVIDE","AMPERSAND","BACKSLASH","EQ","EQEQ","NE","LT","LE",
    "GT","GE","COMMA","TILDE","OR","LAND","LOR","IN_CAP","OUT_CAP",
    "INOUT_CAP","KW_IN","KW_AND","KW_BOOL","KW_INT","KW_FLOAT",
    "KW_STRING","KW_WITH","KW_LOCATION","KW_WSDL","KW_NIL","KW_SCHEMA",
    "KW_LET","KW_NEW","KW_JOIN","KW_SPAWN","KW_MATCH","KW_WHEN",
    "KW_IMPORT","KW_FALSE","KW_TRUE","KW_IF","KW_THEN","KW_ELSE",
    "Selectdec","UMINUS",
  };

  /** index-checked interface to yyNames[].
      @param token single character or %token value.
      @return token name or [illegal] or [unknown].
    */
  public static string yyname (int token) {
    if ((token < 0) || (token > yyNames.Length)) return "[illegal]";
    string name;
    if ((name = yyNames[token]) != null) return name;
    return "[unknown]";
  }

  /** computes list of expected tokens on error by tracing the tables.
      @param state for which to compute the list.
      @return list of token names.
    */
  protected string[] yyExpecting (int state) {
    int token, n, len = 0;
    bool[] ok = new bool[yyNames.Length];

    if ((n = yySindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    if ((n = yyRindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }

    string [] result = new string[len];
    for (n = token = 0; n < len;  ++ token)
      if (ok[token]) result[n++] = yyNames[token];
    return result;
  }

  /** the generated parser, with debugging messages.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @param yydebug debug message writer implementing yyDebug, or null.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex, Object yyd)
				 {
//t    this.debug = (yydebug.yyDebug)yyd;
    return yyparse(yyLex);
  }

  /** initial size and increment of the state/value stack [default 256].
      This is not final so that it can be overwritten outside of invocations
      of yyparse().
    */
  protected int yyMax;

  /** executed at the beginning of a reduce action.
      Used as $$ = yyDefault($1), prior to the user-specified action, if any.
      Can be overwritten to provide deep copy, etc.
      @param first value for $1, or null.
      @return first.
    */
  protected Object yyDefault (Object first) {
    return first;
  }

  /** the generated parser.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex)
				{
    if (yyMax <= 0) yyMax = 256;			// initial size
    int yyState = 0;                                   // state stack ptr
    int [] yyStates = new int[yyMax];	                // state stack 
    Object yyVal = null;                               // value stack ptr
    Object [] yyVals = new Object[yyMax];	        // value stack
    int yyToken = -1;					// current input
    int yyErrorFlag = 0;				// #tks to shift

    int yyTop = 0;
    goto skip;
    yyLoop:
    yyTop++;
    skip:
    for (;; ++ yyTop) {
      if (yyTop >= yyStates.Length) {			// dynamically increase
        int[] i = new int[yyStates.Length+yyMax];
        yyStates.CopyTo (i, 0);
        yyStates = i;
        Object[] o = new Object[yyVals.Length+yyMax];
        yyVals.CopyTo (o, 0);
        yyVals = o;
      }
      yyStates[yyTop] = yyState;
      yyVals[yyTop] = yyVal;
//t      if (debug != null) debug.push(yyState, yyVal);

      yyDiscarded: for (;;) {	// discarding a token does not change stack
        int yyN;
        if ((yyN = yyDefRed[yyState]) == 0) {	// else [default] reduce (yyN)
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t              debug.lex(yyState, yyToken, yyname(yyToken), yyLex.value());
          }
          if ((yyN = yySindex[yyState]) != 0 && ((yyN += yyToken) >= 0)
              && (yyN < yyTable.Length) && (yyCheck[yyN] == yyToken)) {
//t            if (debug != null)
//t              debug.shift(yyState, yyTable[yyN], yyErrorFlag-1);
            yyState = yyTable[yyN];		// shift to yyN
            yyVal = yyLex.value();
            yyToken = -1;
            if (yyErrorFlag > 0) -- yyErrorFlag;
            goto yyLoop;
          }
          if ((yyN = yyRindex[yyState]) != 0 && (yyN += yyToken) >= 0
              && yyN < yyTable.Length && yyCheck[yyN] == yyToken)
            yyN = yyTable[yyN];			// reduce (yyN)
          else
            switch (yyErrorFlag) {
  
            case 0:
              yyerror(String.Format ("syntax error, got token `{0}'", yyname (yyToken)), yyExpecting(yyState));
//t              if (debug != null) debug.error("syntax error");
              goto case 1;
            case 1: case 2:
              yyErrorFlag = 3;
              do {
                if ((yyN = yySindex[yyStates[yyTop]]) != 0
                    && (yyN += Token.yyErrorCode) >= 0 && yyN < yyTable.Length
                    && yyCheck[yyN] == Token.yyErrorCode) {
//t                  if (debug != null)
//t                    debug.shift(yyStates[yyTop], yyTable[yyN], 3);
                  yyState = yyTable[yyN];
                  yyVal = yyLex.value();
                  goto yyLoop;
                }
//t                if (debug != null) debug.pop(yyStates[yyTop]);
              } while (-- yyTop >= 0);
//t              if (debug != null) debug.reject();
              throw new yyParser.yyException("irrecoverable syntax error");
  
            case 3:
              if (yyToken == 0) {
//t                if (debug != null) debug.reject();
                throw new yyParser.yyException("irrecoverable syntax error at end-of-file");
              }
//t              if (debug != null)
//t                debug.discard(yyState, yyToken, yyname(yyToken),
//t  							yyLex.value());
              yyToken = -1;
              goto yyDiscarded;		// leave stack alone
            }
        }
        int yyV = yyTop + 1-yyLen[yyN];
//t        if (debug != null)
//t          debug.reduce(yyState, yyStates[yyV-1], yyN, yyRule[yyN], yyLen[yyN]);
        yyVal = yyDefault(yyV > yyTop ? null : yyVals[yyV]);
        switch (yyN) {
case 1:
#line 198 "Parser.jay"
  {
		yyVal = new NilProcess();
		SetLocation(yyVal, ((GenericValue)yyVals[0+yyTop]).Location);
	}
  break;
case 2:
#line 204 "Parser.jay"
  {
		yyVal = new OutputProcess(((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 3:
#line 210 "Parser.jay"
  {
		yyVal = new InputProcess(((IExpression)yyVals[-5+yyTop]), ((bool)yyVals[-3+yyTop]), ((IPattern)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-5+yyTop])));
	}
  break;
case 4:
#line 216 "Parser.jay"
  {
		yyVal = new SchemaProcess(((IList_SchemaProcess_Declaration)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-3+yyTop]).Location);
	}
  break;
case 5:
#line 222 "Parser.jay"
  {
		MatchProcess.Branch branch = new MatchProcess.Branch(((IPattern)yyVals[-4+yyTop]), null, ((IProcess)yyVals[0+yyTop]));
		SetLocation(branch, GetLocation(((IPattern)yyVals[-4+yyTop])));
	 	IList<MatchProcess.Branch> branches = new List<MatchProcess.Branch>();
		branches.Add(branch);
	  	yyVal = new MatchProcess(((IExpression)yyVals[-2+yyTop]), branches);
		SetLocation(yyVal, ((GenericValue)yyVals[-5+yyTop]).Location);
	}
  break;
case 6:
#line 232 "Parser.jay"
  {
		yyVal = new NewProcess(((StringValue)yyVals[-4+yyTop]).Value, ((IDictionary_FieldDecList)yyVals[-3+yyTop]), ((string)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-5+yyTop]).Location);
	}
  break;
case 7:
#line 238 "Parser.jay"
  {
		yyVal = new ImportProcess(((StringValue)yyVals[-4+yyTop]).Value, ((IDictionary_FieldDecList)yyVals[-3+yyTop]), ((string)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-5+yyTop]).Location);
	}
  break;
case 8:
#line 244 "Parser.jay"
  {
		yyVal = new SpawnProcess(((IProcess)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-4+yyTop]).Location);
	}
  break;
case 9:
#line 250 "Parser.jay"
  {
		yyVal = new MatchProcess(((IExpression)yyVals[-4+yyTop]), ((IList_MatchProcess_Branch)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-5+yyTop]).Location);
	}
  break;
case 10:
#line 256 "Parser.jay"
  {
		yyVal = new JoinProcess(((bool)yyVals[-4+yyTop]), ((IList_JoinPatternRow)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-3+yyTop]).Location);
	}
  break;
case 11:
#line 262 "Parser.jay"
  {
		IList<MatchProcess.Branch> branches = new List<MatchProcess.Branch>();
		MatchProcess.Branch trueBranch = new MatchProcess.Branch(new BasicSchemaLiteralSchema(new BoolLiteral(true)), null, ((IProcess)yyVals[-1+yyTop]));
		MatchProcess.Branch falseBranch = new MatchProcess.Branch(new BasicSchema(new BoolType()), null, ((IProcess)yyVals[0+yyTop]));
		SetLocation(trueBranch, ((GenericValue)yyVals[-2+yyTop]).Location);
		SetLocation(falseBranch, GetLocation(((IProcess)yyVals[-1+yyTop])));
		branches.Add(trueBranch);
		branches.Add(falseBranch);
		yyVal = new MatchProcess(((IExpression)yyVals[-3+yyTop]), branches);
		SetLocation(yyVal, ((GenericValue)yyVals[-4+yyTop]).Location);
	}
  break;
case 12:
#line 275 "Parser.jay"
  { yyVal = ((IProcess)yyVals[-1+yyTop]); }
  break;
case 13:
#line 280 "Parser.jay"
  { yyVal = new NilProcess(); }
  break;
case 14:
#line 283 "Parser.jay"
  { yyVal = ((IProcess)yyVals[0+yyTop]); }
  break;
case 15:
#line 288 "Parser.jay"
  { yyVal = ((IDictionary_FieldDecList)yyVals[-1+yyTop]); }
  break;
case 16:
#line 291 "Parser.jay"
  {	
		IDictionary<string, ISchema> operations = new Dictionary<string, ISchema>();
	  operations.Add("default", ((ISchema)yyVals[0+yyTop]));
		yyVal = operations;
	}
  break;
case 17:
#line 300 "Parser.jay"
  { yyVal = new Dictionary<string, ISchema>(); }
  break;
case 18:
#line 303 "Parser.jay"
  { yyVal = ((IDictionary_FieldDecList)yyVals[0+yyTop]); }
  break;
case 19:
#line 308 "Parser.jay"
  { 		
		IDictionary<string, ISchema> fields = new Dictionary<string, ISchema>();
		fields.Add(((StringValue)yyVals[-2+yyTop]).Value, ((ISchema)yyVals[0+yyTop]));
		yyVal = fields;
	}
  break;
case 20:
#line 315 "Parser.jay"
  { 		
		IDictionary<string, ISchema> fields = (Dictionary<string, ISchema>) ((IDictionary_FieldDecList)yyVals[0+yyTop]);
	 	if (fields.ContainsKey(((StringValue)yyVals[-4+yyTop]).Value))
		  CompilerOutput.Warning("duplicated field " + ((StringValue)yyVals[-4+yyTop]).Value + " (ignored)", ((StringValue)yyVals[-4+yyTop]).Location);
		else
		  fields.Add(((StringValue)yyVals[-4+yyTop]).Value, ((ISchema)yyVals[-2+yyTop]));
	 	yyVal = fields; 	
	}
  break;
case 21:
#line 327 "Parser.jay"
  { yyVal = false; }
  break;
case 22:
#line 330 "Parser.jay"
  { yyVal = true; }
  break;
case 23:
#line 335 "Parser.jay"
  {	
		List<JoinPatternRow> rows = new List<JoinPatternRow>();
		rows.Add(((JoinPatternRow)yyVals[0+yyTop]));
		yyVal = rows;
	}
  break;
case 24:
#line 342 "Parser.jay"
  {
	  IList<JoinPatternRow> rows = (IList<JoinPatternRow>) ((IList_JoinPatternRow)yyVals[-2+yyTop]);
		rows.Add(((JoinPatternRow)yyVals[0+yyTop]));
		yyVal = rows;
	}
  break;
case 25:
#line 351 "Parser.jay"
  {		
		yyVal = new JoinPatternRow(((IList_JoinPatternAtom)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-1+yyTop]).Location); /* TODO: IMPROVE THIS LOCATION*/
	}
  break;
case 26:
#line 359 "Parser.jay"
  {
		IList<JoinPatternAtom> atoms = new List<JoinPatternAtom>();
		atoms.Add(((JoinPatternAtom)yyVals[0+yyTop]));
		yyVal = atoms;
	}
  break;
case 27:
#line 366 "Parser.jay"
  {
		IList<JoinPatternAtom> atoms = (IList<JoinPatternAtom>) ((IList_JoinPatternAtom)yyVals[-2+yyTop]);
		atoms.Add(((JoinPatternAtom)yyVals[0+yyTop]));
		yyVal = atoms;
	}
  break;
case 28:
#line 375 "Parser.jay"
  {
		yyVal = new JoinPatternAtom(((IExpression)yyVals[-2+yyTop]), ((IPattern)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 29:
#line 383 "Parser.jay"
  {
		IList<SchemaProcess.Declaration> decls = new List<SchemaProcess.Declaration>();
		decls.Add(((SchemaProcess.Declaration)yyVals[0+yyTop]));
		yyVal = decls;
	}
  break;
case 30:
#line 390 "Parser.jay"
  {
		IList<SchemaProcess.Declaration> decls = (IList<SchemaProcess.Declaration>) ((IList_SchemaProcess_Declaration)yyVals[-2+yyTop]);
		decls.Add(((SchemaProcess.Declaration)yyVals[0+yyTop]));
		yyVal = decls;
	}
  break;
case 31:
#line 399 "Parser.jay"
  {
		yyVal = new SchemaProcess.Declaration(((StringValue)yyVals[-2+yyTop]).Value, ((ISchema)yyVals[0+yyTop]));
		SetLocation(yyVal, ((StringValue)yyVals[-2+yyTop]).Location);
	}
  break;
case 32:
#line 407 "Parser.jay"
  {
	  yyVal = new List<MatchProcess.Branch>();
		((List<MatchProcess.Branch>) yyVal).Add(((MatchProcess.Branch)yyVals[0+yyTop]));
	}
  break;
case 33:
#line 413 "Parser.jay"
  {
		((IList_MatchProcess_Branch)yyVals[-2+yyTop]).Add(((MatchProcess.Branch)yyVals[0+yyTop]));
		yyVal = ((IList_MatchProcess_Branch)yyVals[-2+yyTop]);
	}
  break;
case 34:
#line 421 "Parser.jay"
  {
		yyVal = new MatchProcess.Branch(((IPattern)yyVals[-3+yyTop]), ((IExpression)yyVals[-2+yyTop]), ((IProcess)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IPattern)yyVals[-3+yyTop])));
	}
  break;
case 35:
#line 429 "Parser.jay"
  { yyVal = null; }
  break;
case 36:
#line 432 "Parser.jay"
  {
		yyVal = ((IExpression)yyVals[0+yyTop]);
		SetLocation(yyVal, ((GenericValue)yyVals[-1+yyTop]).Location);
	}
  break;
case 37:
#line 440 "Parser.jay"
  {
		yyVal = new VoidSchema();
		SetLocation(yyVal, lexer.location);
	}
  break;
case 38:
#line 446 "Parser.jay"
  { yyVal = ((IPattern)yyVals[0+yyTop]); }
  break;
case 39:
#line 451 "Parser.jay"
  {
		yyVal = new VoidExpr();
		SetLocation(yyVal, lexer.location);
	}
  break;
case 40:
#line 457 "Parser.jay"
  { yyVal = ((IExpression)yyVals[0+yyTop]); }
  break;
case 41:
#line 462 "Parser.jay"
  {
		yyVal = new IntLiteral(((IntValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((IntValue)yyVals[0+yyTop]).Location);
	}
  break;
case 42:
#line 468 "Parser.jay"
  {
		yyVal = new FloatLiteral(((FloatValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((FloatValue)yyVals[0+yyTop]).Location);
	}
  break;
case 43:
#line 474 "Parser.jay"
  {
		yyVal = new StringLiteral(((StringValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((StringValue)yyVals[0+yyTop]).Location);
	}
  break;
case 44:
#line 480 "Parser.jay"
  {
		yyVal = new BoolLiteral(false);
		SetLocation(yyVal, ((GenericValue)yyVals[0+yyTop]).Location);
	}
  break;
case 45:
#line 486 "Parser.jay"
  {
		yyVal = new BoolLiteral(true);
		SetLocation(yyVal, ((GenericValue)yyVals[0+yyTop]).Location);
	}
  break;
case 46:
#line 494 "Parser.jay"
  { yyVal = new Dictionary<string, IExpression>(); }
  break;
case 47:
#line 497 "Parser.jay"
  { yyVal = ((IDictionary_FieldDefList)yyVals[0+yyTop]); }
  break;
case 48:
#line 502 "Parser.jay"
  {
		IDictionary<string, IExpression> fields = new Dictionary<string, IExpression>();
		fields.Add(((StringValue)yyVals[-2+yyTop]).Value, ((IExpression)yyVals[0+yyTop]));
		yyVal = fields;
	}
  break;
case 49:
#line 509 "Parser.jay"
  {
		IDictionary<string, IExpression> fields = (IDictionary<string, IExpression>) ((IDictionary_FieldDefList)yyVals[0+yyTop]);
	 	if (fields.ContainsKey(((StringValue)yyVals[-4+yyTop]).Value))
		  CompilerOutput.Warning("duplicated field " + ((StringValue)yyVals[-4+yyTop]).Value + " (ignored)", ((StringValue)yyVals[-4+yyTop]).Location);
		else
		  fields.Add(((StringValue)yyVals[-4+yyTop]).Value, ((IExpression)yyVals[-2+yyTop]));
		yyVal = fields;
	}
  break;
case 50:
#line 521 "Parser.jay"
  { yyVal = new RecordExpr(); }
  break;
case 51:
#line 524 "Parser.jay"
  { yyVal = new RecordExpr(((IDictionary_FieldDefList)yyVals[-1+yyTop])); }
  break;
case 52:
#line 529 "Parser.jay"
  { yyVal = ((IExpression)yyVals[-1+yyTop]); }
  break;
case 53:
#line 532 "Parser.jay"
  {
		yyVal = new VariableExpr(((StringValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((StringValue)yyVals[0+yyTop]).Location);
	}
  break;
case 54:
#line 538 "Parser.jay"
  {
		yyVal = new SelectExpr(((IExpression)yyVals[-2+yyTop]), ((StringValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 55:
#line 545 "Parser.jay"
  {
		yyVal = new RecordExpr(((IDictionary_FieldDefList)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-2+yyTop]).Location);
	}
  break;
case 56:
#line 551 "Parser.jay"
  {
		yyVal = new LiteralExpr(((IBasicTypeLiteral)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IBasicTypeLiteral)yyVals[0+yyTop])));
	}
  break;
case 57:
#line 557 "Parser.jay"
  {
		yyVal = new LabelledExpr(((StringValue)yyVals[-4+yyTop]).Value, ((IExpression)yyVals[-1+yyTop]), ((RecordExpr)yyVals[-3+yyTop]));
		SetLocation(yyVal, ((StringValue)yyVals[-4+yyTop]).Location);
	}
  break;
case 58:
#line 563 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.PLUS, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 59:
#line 569 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.MINUS, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 60:
#line 575 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.TIMES, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 61:
#line 581 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.DIVIDE, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 62:
#line 587 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.LAND, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 63:
#line 593 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.LOR, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 64:
#line 599 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.EQ, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 65:
#line 605 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.NE, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 66:
#line 611 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.GT, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 67:
#line 617 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.GE, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 68:
#line 623 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.LT, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 69:
#line 629 "Parser.jay"
  {
		yyVal = new BinaryOpExpr(BinaryOp.LE, ((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 70:
#line 635 "Parser.jay"
  {
		yyVal = new SequenceExpr(((IExpression)yyVals[-2+yyTop]), ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IExpression)yyVals[-2+yyTop])));
	}
  break;
case 71:
#line 641 "Parser.jay"
  {
		yyVal = new UnaryOpExpr(UnaryOp.LNOT, ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-1+yyTop]).Location);
	}
  break;
case 72:
#line 647 "Parser.jay"
  {
		yyVal = new UnaryOpExpr(UnaryOp.UMINUS, ((IExpression)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-1+yyTop]).Location);
	}
  break;
case 73:
#line 655 "Parser.jay"
  {
		yyVal = LabelSet.Singleton(((StringValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((StringValue)yyVals[0+yyTop]).Location);
	}
  break;
case 74:
#line 661 "Parser.jay"
  {
		yyVal = LabelSet.Any();
		SetLocation(yyVal, ((GenericValue)yyVals[0+yyTop]).Location);
	}
  break;
case 75:
#line 667 "Parser.jay"
  {
		yyVal = LabelSet.With((ISet<string>)((ISet_string)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-2+yyTop]).Location);
	}
  break;
case 76:
#line 673 "Parser.jay"
  {
		yyVal = LabelSet.Without((ISet<string>)((ISet_string)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-4+yyTop]).Location);
	}
  break;
case 77:
#line 681 "Parser.jay"
  { yyVal = new ArraySet<string>(); }
  break;
case 78:
#line 684 "Parser.jay"
  { yyVal = ((ISet_string)yyVals[0+yyTop]); }
  break;
case 79:
#line 689 "Parser.jay"
  {
		ISet<string> labels = new ArraySet<string>();
		labels.Add(((StringValue)yyVals[0+yyTop]).Value);
		yyVal = labels;
	}
  break;
case 80:
#line 696 "Parser.jay"
  {
		Debug.Assert(yyVal is ISet<string>);
		ISet<string> labels = (ArraySet<string>) yyVal;
		labels.Add(((StringValue)yyVals[0+yyTop]).Value);
	}
  break;
case 81:
#line 705 "Parser.jay"
  {
		yyVal = new VoidSchema();
		SetLocation(yyVal, lexer.location);
	}
  break;
case 82:
#line 711 "Parser.jay"
  { yyVal = ((ISchema)yyVals[0+yyTop]); }
  break;
case 83:
#line 716 "Parser.jay"
  {
		yyVal = new BoolType();
	}
  break;
case 84:
#line 721 "Parser.jay"
  {
		yyVal = new IntType();
	}
  break;
case 85:
#line 726 "Parser.jay"
  {
		yyVal = new FloatType();
	}
  break;
case 86:
#line 731 "Parser.jay"
  {
		yyVal = new StringType();
	}
  break;
case 87:
#line 738 "Parser.jay"
  { yyVal = new RecordSchema(); }
  break;
case 88:
#line 741 "Parser.jay"
  { yyVal = new RecordSchema(((IDictionary_FieldDecList)yyVals[-1+yyTop])); }
  break;
case 89:
#line 746 "Parser.jay"
  {
		yyVal = new BasicSchema(((IBasicType)yyVals[0+yyTop]));
		SetLocation(yyVal, lexer.location);
	}
  break;
case 90:
#line 752 "Parser.jay"
  {
		yyVal = new BasicSchemaLiteralSchema(((IBasicTypeLiteral)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((IBasicTypeLiteral)yyVals[0+yyTop])));
	}
  break;
case 91:
#line 758 "Parser.jay"
  {
		yyVal = new ChannelSchema(((ISchema)yyVals[-2+yyTop]), ((ChannelType.CAPABILITY)yyVals[0+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-3+yyTop]).Location);
	}
  break;
case 92:
#line 764 "Parser.jay"
  {
		yyVal = new FunctionSchema(((ISchema)yyVals[-3+yyTop]), ((ISchema)yyVals[0+yyTop]), ((ChannelType.CAPABILITY)yyVals[-1+yyTop]));
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-3+yyTop])));
	}
  break;
case 93:
#line 770 "Parser.jay"
  {
		yyVal = UnionSchema.MakeNew(((ISchema)yyVals[-2+yyTop]), ((ISchema)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-2+yyTop])));
	}
  break;
case 94:
#line 776 "Parser.jay"
  {
		yyVal = new ConstantSchema(((StringValue)yyVals[0+yyTop]).Value);
		SetLocation(yyVal, ((StringValue)yyVals[0+yyTop]).Location);
	}
  break;
case 95:
#line 782 "Parser.jay"
  {
		yyVal = new BindSchema(((StringValue)yyVals[-2+yyTop]).Value, ((ISchema)yyVals[0+yyTop]));
		SetLocation(yyVal, ((StringValue)yyVals[-2+yyTop]).Location);
	}
  break;
case 96:
#line 788 "Parser.jay"
  {
		yyVal = new VoidSchema();
		SetLocation(yyVal, ((GenericValue)yyVals[-1+yyTop]).Location);
	}
  break;
case 97:
#line 794 "Parser.jay"
  { yyVal = ((ISchema)yyVals[-1+yyTop]); }
  break;
case 98:
#line 797 "Parser.jay"
  {
		yyVal = new LabelledSchema(((ILabelSet)yyVals[-4+yyTop]), ((ISchema)yyVals[-1+yyTop]), ((RecordSchema)yyVals[-3+yyTop]));
		SetLocation(yyVal, GetLocation(((ILabelSet)yyVals[-4+yyTop])));
	}
  break;
case 99:
#line 803 "Parser.jay"
  {
	  yyVal = new RecordSchema(((IDictionary_FieldDecList)yyVals[-1+yyTop]));
		SetLocation(yyVal, ((GenericValue)yyVals[-2+yyTop]).Location);
	}
  break;
case 100:
#line 809 "Parser.jay"
  {
		yyVal = new SequenceSchema(((ISchema)yyVals[-2+yyTop]), ((ISchema)yyVals[0+yyTop]));
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-2+yyTop])));
	}
  break;
case 101:
#line 815 "Parser.jay"
  {
		yyVal = new StarSchema(((ISchema)yyVals[-1+yyTop]));
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-1+yyTop])));
	}
  break;
case 102:
#line 821 "Parser.jay"
  {
		yyVal = new PlusSchema(((ISchema)yyVals[-1+yyTop]));
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-1+yyTop])));
	}
  break;
case 103:
#line 827 "Parser.jay"
  {
		yyVal = new RepetitionSchema(((ISchema)yyVals[-3+yyTop]), (uint) ((IntValue)yyVals[-1+yyTop]).Value, (uint) ((IntValue)yyVals[-1+yyTop]).Value);
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-3+yyTop])));
	}
  break;
case 104:
#line 833 "Parser.jay"
  {
		yyVal = new RepetitionSchema(((ISchema)yyVals[-5+yyTop]), (uint) ((IntValue)yyVals[-3+yyTop]).Value, (uint) ((IntValue)yyVals[-1+yyTop]).Value);
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-5+yyTop])));
	}
  break;
case 105:
#line 839 "Parser.jay"
  {
		yyVal = new RepetitionSchema(((ISchema)yyVals[-1+yyTop]), 0, 1);
		SetLocation(yyVal, GetLocation(((ISchema)yyVals[-1+yyTop])));
	}
  break;
case 106:
#line 847 "Parser.jay"
  { yyVal = ChannelType.CAPABILITY.IN; }
  break;
case 107:
#line 850 "Parser.jay"
  { yyVal = ChannelType.CAPABILITY.OUT; }
  break;
case 108:
#line 853 "Parser.jay"
  { yyVal = ChannelType.CAPABILITY.INOUT; }
  break;
case 109:
#line 858 "Parser.jay"
  { yyVal = ChannelType.CAPABILITY.OUT; }
  break;
case 110:
#line 861 "Parser.jay"
  { yyVal = ((ChannelType.CAPABILITY)yyVals[0+yyTop]); }
  break;
case 111:
#line 866 "Parser.jay"
  { yyVal = ((StringValue)yyVals[0+yyTop]).Value; }
  break;
case 112:
#line 870 "Parser.jay"
  { yyVal = ""; }
  break;
case 113:
#line 873 "Parser.jay"
  { yyVal = ((string)yyVals[0+yyTop]); }
  break;
case 114:
#line 878 "Parser.jay"
  { yyVal = ((ISchema)yyVals[0+yyTop]); }
  break;
#line default
        }
        yyTop -= yyLen[yyN];
        yyState = yyStates[yyTop];
        int yyM = yyLhs[yyN];
        if (yyState == 0 && yyM == 0) {
//t          if (debug != null) debug.shift(0, yyFinal);
          yyState = yyFinal;
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t               debug.lex(yyState, yyToken,yyname(yyToken), yyLex.value());
          }
          if (yyToken == 0) {
//t            if (debug != null) debug.accept(yyVal);
            return yyVal;
          }
          goto yyLoop;
        }
        if (((yyN = yyGindex[yyM]) != 0) && ((yyN += yyState) >= 0)
            && (yyN < yyTable.Length) && (yyCheck[yyN] == yyState))
          yyState = yyTable[yyN];
        else
          yyState = yyDgoto[yyM];
//t        if (debug != null) debug.shift(yyStates[yyTop], yyState);
	 goto yyLoop;
      }
    }
  }

   static  short [] yyLhs  = {              -1,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    1,    1,   31,   31,   30,   30,   29,   29,
    3,    3,   23,   23,   26,   24,   24,   27,   25,   25,
   28,    4,    4,    5,    2,    2,    7,    7,   10,   10,
    8,    8,    8,    8,    8,   33,   33,   32,   32,   11,
   11,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,    9,    9,    9,    9,    9,    9,    9,    9,
    9,    9,   16,   16,   16,   16,   17,   17,   18,   18,
   14,   14,   12,   12,   12,   12,   15,   15,   13,   13,
   13,   13,   13,   13,   13,   13,   13,   13,   13,   13,
   13,   13,   13,   13,   13,   19,   19,   19,   20,   20,
   21,   22,   22,    6,
  };
   static  short [] yyLen = {           2,
    1,    3,    6,    4,    6,    6,    6,    5,    6,    5,
    5,    3,    0,    2,    3,    2,    0,    1,    3,    5,
    0,    1,    1,    3,    3,    1,    3,    3,    1,    3,
    3,    1,    3,    4,    0,    2,    0,    1,    0,    1,
    1,    1,    1,    1,    1,    0,    1,    3,    5,    0,
    3,    3,    1,    3,    3,    1,    5,    3,    3,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
    2,    2,    1,    1,    3,    5,    0,    1,    1,    3,
    0,    1,    1,    1,    1,    1,    0,    3,    1,    1,
    4,    4,    3,    1,    3,    2,    3,    5,    3,    3,
    2,    2,    4,    6,    2,    1,    1,    1,    0,    1,
    3,    0,    1,    1,
  };
   static  short [] yyDefRed = {            0,
   43,   41,   42,    0,    0,    0,    0,    0,   22,    1,
    0,    0,    0,    0,    0,    0,   44,   45,    0,    0,
    0,   56,    0,    0,    0,    0,    0,   47,    0,    0,
    0,    0,   71,   72,    0,    0,   29,    0,    0,    0,
    0,   74,   83,   84,   85,   86,    0,   90,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   12,   55,   52,
    0,    0,    0,    0,    0,    0,    0,    0,   18,    0,
   96,    0,    0,    0,    0,    0,    0,  105,  102,  101,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   54,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   51,    0,
    0,    0,    4,   30,    0,    0,    0,   75,    0,   99,
   97,    0,    0,  106,  107,  108,  110,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  113,    0,    0,
    0,    0,    0,    0,    0,    0,   23,   26,    0,   57,
    0,    0,   79,    0,   80,   91,    0,    0,  103,    0,
   88,    0,   15,    0,    0,    8,    0,   32,    0,    0,
    0,   11,    0,   10,    0,    0,    0,    0,   49,    0,
   76,    5,    0,   98,  111,    6,    9,    0,    0,    0,
    7,   14,   38,   28,   24,   25,   27,    3,   20,  104,
   33,    0,    0,   34,
  };
  protected static  short [] yyDgoto  = {            20,
  192,  210,   21,  187,  188,  189,  214,   22,   23,   32,
   25,   49,   50,   94,  104,   51,   87,   88,  147,  148,
  158,  159,  165,  166,   36,  167,  168,   37,   89,   90,
  107,   28,   29,
  };
  protected static  short [] yySindex = {         -224,
    0,    0,    0, -249,  -57, -135, -135, -135,    0,    0,
 -217,  539, -176, -170, -135, -173,    0,    0, -135,    0,
 -211,    0,  861, -159, -160, -253, -154,    0, -153, -159,
  942, -156,    0,    0, -166, -237,    0, -152, -259,  520,
  539,    0,    0,    0,    0,    0, -163,    0,    0,  184,
 -145, -214, -224,  832, -214,  813, -142, -150, -135, -131,
 -135, -135, -135, -135, -135, -135, -135, -135, -135, -135,
 -135, -135, -135, -149, -132, -135, -135,    0,    0,    0,
  539, -224, -217,  539, -138, -143, -130, -148,    0, -127,
    0, -103,  184, -147, -135, -269, -116,    0,    0,    0,
  539,  539, -115, -121, -115,  539, -151, -111, -108, -151,
 -224, -135,  539,  942,    0, -171, -171, -125, -125,   38,
   38,   38,   38,   38,   38,  466,  466,  466,    0, -109,
  923,  184,    0,    0, -120,  539,  -99,    0,  -90,    0,
    0, -269,  880,    0,    0,    0,    0,  539, -247, -120,
 -120, -138,  -92,  539,  -88,  184,  -97,    0, -118, -224,
  539, -114, -128,  903, -258, -230,    0,    0,  -77,    0,
 -159,  443,    0, -245,    0,    0, -224,   32,    0,  -68,
    0,  -74,    0,  -66, -224,    0, -251,    0, -107, -224,
 -224,    0,  539,    0, -135, -224, -135, -224,    0, -115,
    0,    0,  -67,    0,    0,    0,    0,  539, -135,  -53,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  942, -224,    0,
  };
  protected static  short [] yyRindex = {          -96,
    0,    0,    0,    1, -261,  -61,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  -50,    0,  -46,    0,    0,    0,  -50,
 -174,    0,    0,    0,    0,    0,    0,  628,  -49,    0,
  -69,    0,    0,    0,    0,    0,    0,    0,    0, -250,
  -42,    0,  -96,    0,    0,    0,    0,  566,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  -36,    0,    0,    0,    0,
    0,  -96,    0,    0, -243,    0,    0,  -31,    0,    0,
    0,    0, -263,    0,    0,  585,    0,    0,    0,    0,
    0,    0,  -49,    0,    0,    0,  -52,    0,    0,  -52,
  -96,    0,    0,   46,    0,  153,  193,   57,  113,  227,
  261,  295,  329,  363,  406,  393,  465,  499,    0,    0,
  -30, -220,    0,    0,  727,    0,    0,    0,    0,    0,
    0,  672,    0,    0,    0,    0,    0,    0,    0,  747,
  782,    0,    0,  -26,    0, -284,    0,    0,    0,  -96,
    0,    0,   16,    0,    0,    0,    0,    0,    0,    0,
    0,  -18,    0,    0,    0,    0,  -96,  692,    0,    0,
    0,    0,    0,    0,  -96,    0,    0,    0,  -11,  -96,
  -96,    0, -225,    0,    0,  -96,    0,  -96,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   -8,  -96,    0,
  };
  protected static  short [] yyGindex = {           -2,
    0,    0,  202,    0,   54,  -12,    0,   15,    2,  187,
    0,    0,   64,  111,    0,    0,    0,  130,    0,  131,
    0,  172,    0,    0,    0,   95,  101,  216, -100,  197,
  246,  133,  281,
  };
  protected static  short [] yyTable = {            47,
   53,   85,   27,   46,  155,   82,  194,   31,   33,   34,
   24,   16,  114,  207,   24,   13,   54,  179,   16,  201,
   56,   79,   82,  144,  145,  146,   48,   77,  114,   86,
  114,  195,  196,    1,    2,    3,    4,   37,  208,    5,
  180,    6,  139,   35,   79,    2,    7,   21,  197,  105,
  108,    8,    9,   37,   48,   48,   60,  106,   82,   83,
  114,  114,  116,  117,  118,  119,  120,  121,  122,  123,
  124,  125,  126,  127,  128,   31,   31,   31,  131,  133,
   10,   11,   12,   13,   52,   14,   15,   55,   16,   17,
   18,   19,   40,   53,   40,   48,  143,   57,   48,  219,
  169,   74,   60,   92,   93,   63,   64,   76,  163,   78,
   80,   79,   61,  164,   81,   48,   48,   95,  103,   84,
   48,  112,    1,    2,    3,    4,    9,   48,   30,  115,
    6,   77,  129,  136,  138,    7,  137,  140,  142,  139,
    8,   96,  149,   97,  132,  152,  154,  135,   60,   98,
   48,  157,   58,  160,   99,  161,  100,  186,   96,  170,
   97,  173,   48,  141,  150,  151,   98,  101,   48,  156,
  175,   99,  181,  100,  202,   48,  183,  185,   17,   18,
  213,  190,  206,  184,  101,  198,  102,  211,  212,  191,
  203,  205,   59,  216,  204,  218,  164,  220,  164,  172,
    1,    2,    3,   26,  209,   39,    5,   48,    6,  223,
  222,  178,   21,    7,   46,   17,   81,   93,    8,    9,
  224,   50,   48,   53,   53,   87,   64,   53,   53,   53,
   53,   53,   39,   78,   48,   53,   53,   53,   53,   53,
   53,   53,   81,  112,   53,   53,   19,   10,   11,   12,
   13,   35,   14,   15,   36,   16,   17,   18,   19,  113,
   65,  221,  130,   53,  182,   53,  174,   53,   50,   53,
   53,   53,  176,   53,   53,   53,   53,   53,   53,   13,
   13,  162,   53,   53,   53,   53,   53,   53,   53,  215,
   53,   53,   53,   96,   68,   97,   53,  217,  134,  153,
  110,   98,   53,  199,   75,   13,   99,    0,  100,    2,
    2,   60,   61,   62,   63,   64,    0,   53,   53,   60,
   60,   60,    0,   60,    0,   60,   60,   60,   69,   60,
    0,   60,   60,   60,   60,    2,    0,    0,   60,   60,
   60,   60,   60,   60,   60,    0,   60,   60,   60,    0,
    0,    0,   60,    0,    0,    0,    0,    0,   60,    0,
    0,    0,   66,    2,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   60,   60,   61,   61,   61,    0,   61,
    0,   61,   61,   61,    0,   61,    0,   61,   61,   61,
   61,    0,   70,    0,   61,   61,   61,   61,   61,   61,
   61,    0,   61,   61,   61,   67,    0,    0,   61,    0,
    0,    0,    0,    0,   61,   58,   58,   58,    0,   58,
    0,   58,   58,   58,    0,   58,    0,   58,   58,   61,
   61,    0,    0,    0,   58,   58,   58,   58,   58,   58,
   58,    0,   58,   58,   58,   96,    0,   97,   58,    0,
    0,    0,    0,   98,   58,   59,   59,   59,   99,   59,
  100,   59,   59,   59,   62,   59,    0,   59,   59,   58,
   58,  101,    0,  102,   59,   59,   59,   59,   59,   59,
   59,    0,   59,   59,   59,    0,    0,    0,   59,   64,
   64,   64,    0,   64,   59,   64,   64,   64,   63,   64,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   59,
   59,    0,    0,    0,   64,    0,   64,   64,   64,    0,
    0,    0,   64,   65,   65,   65,    0,   65,   64,   65,
   65,   65,    0,   65,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   64,   64,    0,    0,    0,   65,    0,
   65,   65,   65,    0,    0,    0,   65,   68,   68,   68,
    0,   68,   65,   68,   68,   68,    0,   68,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   65,   65,    0,
    0,    0,   68,    0,   68,   68,   68,    0,    0,    0,
   68,   69,   69,   69,    0,   69,   68,   69,   69,   69,
    0,   69,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   68,   68,    0,    0,    0,   69,    0,   69,   69,
   69,    0,    0,    0,   69,   66,   66,   66,    0,   66,
   69,   66,   66,   66,    0,   66,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   69,   69,    0,    0,    0,
   66,    0,   66,   66,   66,   70,   70,   70,   66,   70,
    0,   70,   70,   70,   66,   70,    0,    0,   67,   67,
   67,    0,   67,    0,   67,   67,   67,    0,   67,   66,
   66,    0,   70,   70,   70,    0,    0,    0,   70,    0,
    0,    0,    0,   67,   70,   67,   67,   67,    0,    0,
    0,   67,    0,    0,   96,    0,   97,   67,    0,   70,
   70,    0,   98,    0,    0,  200,    0,   99,    0,  100,
    0,    0,   67,   67,    0,    0,    0,   62,   62,   62,
  101,   62,  102,   62,   62,   62,    0,   62,    0,   60,
   61,   62,   63,   64,    0,    0,    0,   65,   66,   67,
   68,   69,   70,   71,   62,   62,   62,    0,    0,    0,
   62,   63,   63,   63,    0,   63,   62,   63,   63,   63,
    0,   63,    0,    0,    0,    0,    0,    1,    2,    3,
   38,   62,   62,   39,    0,   40,   91,    0,   63,   63,
   63,    0,    0,    0,   63,    0,    1,    2,    3,   38,
   63,    0,   39,   41,   40,    0,    0,    0,   42,    0,
    0,    0,    0,    0,    0,   63,   63,   43,   44,   45,
   46,    0,   41,   21,   21,   21,   21,   42,    0,   21,
    0,   21,    0,   17,   18,    0,   43,   44,   45,   46,
    0,    0,  109,  109,  109,  109,    0,    0,  109,   21,
  109,    0,   17,   18,   21,    0,    0,    0,    0,    0,
    0,    0,    0,   21,   21,   21,   21,    0,  109,    0,
    0,    0,    0,  109,    0,    0,    0,    0,    0,   21,
   21,    0,  109,  109,  109,  109,    0,    0,    0,   94,
   94,   73,   94,    0,   94,   73,   94,   94,  109,  109,
   94,    0,   94,    0,   94,    0,   94,    0,   94,    0,
    0,    0,    0,   94,    0,   94,    0,   94,    0,    0,
    0,    0,    0,   94,   94,    0,    0,    0,    0,    0,
   94,    0,    0,  109,  109,  109,  109,    0,  109,   94,
  109,  109,    0,    0,  109,    0,  109,    0,  109,    0,
  109,    0,  109,    0,   92,    0,   92,  109,   92,  109,
   92,  109,    0,    0,   92,    0,    0,  109,  109,    0,
   92,    0,   92,    0,  109,    0,    0,   92,    0,   92,
    0,   92,    0,  109,    0,    0,    0,   92,   92,   95,
    0,   95,    0,   95,   92,   95,    0,    0,    0,   95,
    0,    0,    0,   92,    0,   95,    0,   95,    0,  100,
    0,  100,   95,  100,    0,  100,   95,    0,    0,  100,
    0,    0,   95,   95,    0,  100,    0,  100,    0,   95,
    0,    0,  100,    0,    0,    0,  100,    0,   95,    0,
    0,    0,  100,  100,   93,    0,   93,    0,   93,  100,
   93,    0,    0,    0,   93,    0,    0,    0,  100,    0,
   93,    0,   93,    0,    0,    0,    0,   93,    0,    0,
    0,   93,    0,    0,    0,    0,    0,   93,   93,    0,
    0,    0,    0,    0,   93,    0,   60,   61,   62,   63,
   64,    0,    0,   93,   65,   66,   67,   68,   69,   70,
   71,    0,    0,   72,   73,   60,   61,   62,   63,   64,
    0,    0,    0,   65,   66,   67,   68,   69,   70,   71,
    0,    0,   72,   73,    0,    0,    0,    0,    0,  111,
   58,   59,    0,  109,   60,   61,   62,   63,   64,    0,
    0,    0,   65,   66,   67,   68,   69,   70,   71,    0,
    0,   72,   73,   60,   61,   62,   63,   64,    0,    0,
    0,   65,   66,   67,   68,   69,   70,   71,    0,    0,
   72,   73,  193,    0,    0,  177,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,   67,   68,   69,   70,
   71,    0,    0,   72,   73,  171,   60,   61,   62,   63,
   64,    0,    0,    0,   65,   66,   67,   68,   69,   70,
   71,    0,    0,   72,   73,   60,   61,   62,   63,   64,
    0,    0,    0,   65,   66,   67,   68,   69,   70,   71,
    0,    0,   72,   73,
  };
  protected static  short [] yyCheck = {            12,
    0,  261,    5,  265,  105,  269,  265,    6,    7,    8,
  264,  296,  263,  265,  264,    0,   15,  265,  303,  265,
   19,  265,  286,  293,  294,  295,   12,  281,  279,  289,
  281,  290,  263,  258,  259,  260,  261,  263,  290,  264,
  288,  266,  288,  261,  288,    0,  271,  309,  279,  264,
   53,  276,  277,  279,   40,   41,    0,  272,  296,  297,
   59,  312,   61,   62,   63,   64,   65,   66,   67,   68,
   69,   70,   71,   72,   73,  296,  297,   76,   77,   82,
  305,  306,  307,  308,  261,  310,  311,  261,  313,  314,
  315,  316,  267,  264,  269,   81,   95,  309,   84,  200,
  113,  261,  274,   40,   41,  277,  278,  268,  111,  264,
  267,  265,    0,  112,  281,  101,  102,  281,  264,  272,
  106,  264,  258,  259,  260,  261,  277,  113,  264,  261,
  266,  281,  265,  272,  265,  271,  280,  265,  286,  288,
  276,  262,  259,  264,   81,  261,  268,   84,  274,  270,
  136,  303,    0,  265,  275,  264,  277,  160,  262,  269,
  264,  261,  148,  267,  101,  102,  270,  288,  154,  106,
  261,  275,  265,  277,  177,  161,  265,  296,  314,  315,
  193,  296,  185,  281,  288,  263,  290,  190,  191,  318,
  259,  258,    0,  196,  269,  198,  195,  265,  197,  136,
  258,  259,  260,  261,  312,  267,  264,  193,  266,  263,
  209,  148,  309,  271,  265,  265,  286,  154,  276,  277,
  223,  268,  208,  270,  271,  268,    0,  274,  275,  276,
  277,  278,  269,  265,  265,  282,  283,  284,  285,  286,
  287,  288,  269,  296,  291,  292,  265,  305,  306,  307,
  308,  263,  310,  311,  263,  313,  314,  315,  316,   58,
    0,  208,   76,  263,  154,  265,  137,  267,  268,  269,
  270,  271,  142,  273,  274,  275,  276,  277,  278,  264,
  265,  110,  282,  283,  284,  285,  286,  287,  288,  195,
  290,  291,  292,  262,    0,  264,  296,  197,   83,  103,
   55,  270,  302,  171,   24,  290,  275,   -1,  277,  264,
  265,  274,  275,  276,  277,  278,   -1,  317,  318,  263,
  264,  265,   -1,  267,   -1,  269,  270,  271,    0,  273,
   -1,  275,  276,  277,  278,  290,   -1,   -1,  282,  283,
  284,  285,  286,  287,  288,   -1,  290,  291,  292,   -1,
   -1,   -1,  296,   -1,   -1,   -1,   -1,   -1,  302,   -1,
   -1,   -1,    0,  318,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,  318,  263,  264,  265,   -1,  267,
   -1,  269,  270,  271,   -1,  273,   -1,  275,  276,  277,
  278,   -1,    0,   -1,  282,  283,  284,  285,  286,  287,
  288,   -1,  290,  291,  292,    0,   -1,   -1,  296,   -1,
   -1,   -1,   -1,   -1,  302,  263,  264,  265,   -1,  267,
   -1,  269,  270,  271,   -1,  273,   -1,  275,  276,  317,
  318,   -1,   -1,   -1,  282,  283,  284,  285,  286,  287,
  288,   -1,  290,  291,  292,  262,   -1,  264,  296,   -1,
   -1,   -1,   -1,  270,  302,  263,  264,  265,  275,  267,
  277,  269,  270,  271,    0,  273,   -1,  275,  276,  317,
  318,  288,   -1,  290,  282,  283,  284,  285,  286,  287,
  288,   -1,  290,  291,  292,   -1,   -1,   -1,  296,  263,
  264,  265,   -1,  267,  302,  269,  270,  271,    0,  273,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,
  318,   -1,   -1,   -1,  288,   -1,  290,  291,  292,   -1,
   -1,   -1,  296,  263,  264,  265,   -1,  267,  302,  269,
  270,  271,   -1,  273,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  317,  318,   -1,   -1,   -1,  288,   -1,
  290,  291,  292,   -1,   -1,   -1,  296,  263,  264,  265,
   -1,  267,  302,  269,  270,  271,   -1,  273,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,   -1,
   -1,   -1,  288,   -1,  290,  291,  292,   -1,   -1,   -1,
  296,  263,  264,  265,   -1,  267,  302,  269,  270,  271,
   -1,  273,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  317,  318,   -1,   -1,   -1,  288,   -1,  290,  291,
  292,   -1,   -1,   -1,  296,  263,  264,  265,   -1,  267,
  302,  269,  270,  271,   -1,  273,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  317,  318,   -1,   -1,   -1,
  288,   -1,  290,  291,  292,  263,  264,  265,  296,  267,
   -1,  269,  270,  271,  302,  273,   -1,   -1,  263,  264,
  265,   -1,  267,   -1,  269,  270,  271,   -1,  273,  317,
  318,   -1,  290,  291,  292,   -1,   -1,   -1,  296,   -1,
   -1,   -1,   -1,  288,  302,  290,  291,  292,   -1,   -1,
   -1,  296,   -1,   -1,  262,   -1,  264,  302,   -1,  317,
  318,   -1,  270,   -1,   -1,  273,   -1,  275,   -1,  277,
   -1,   -1,  317,  318,   -1,   -1,   -1,  263,  264,  265,
  288,  267,  290,  269,  270,  271,   -1,  273,   -1,  274,
  275,  276,  277,  278,   -1,   -1,   -1,  282,  283,  284,
  285,  286,  287,  288,  290,  291,  292,   -1,   -1,   -1,
  296,  263,  264,  265,   -1,  267,  302,  269,  270,  271,
   -1,  273,   -1,   -1,   -1,   -1,   -1,  258,  259,  260,
  261,  317,  318,  264,   -1,  266,  267,   -1,  290,  291,
  292,   -1,   -1,   -1,  296,   -1,  258,  259,  260,  261,
  302,   -1,  264,  284,  266,   -1,   -1,   -1,  289,   -1,
   -1,   -1,   -1,   -1,   -1,  317,  318,  298,  299,  300,
  301,   -1,  284,  258,  259,  260,  261,  289,   -1,  264,
   -1,  266,   -1,  314,  315,   -1,  298,  299,  300,  301,
   -1,   -1,  258,  259,  260,  261,   -1,   -1,  264,  284,
  266,   -1,  314,  315,  289,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  298,  299,  300,  301,   -1,  284,   -1,
   -1,   -1,   -1,  289,   -1,   -1,   -1,   -1,   -1,  314,
  315,   -1,  298,  299,  300,  301,   -1,   -1,   -1,  262,
  263,  264,  265,   -1,  267,  268,  269,  270,  314,  315,
  273,   -1,  275,   -1,  277,   -1,  279,   -1,  281,   -1,
   -1,   -1,   -1,  286,   -1,  288,   -1,  290,   -1,   -1,
   -1,   -1,   -1,  296,  297,   -1,   -1,   -1,   -1,   -1,
  303,   -1,   -1,  262,  263,  264,  265,   -1,  267,  312,
  269,  270,   -1,   -1,  273,   -1,  275,   -1,  277,   -1,
  279,   -1,  281,   -1,  263,   -1,  265,  286,  267,  288,
  269,  290,   -1,   -1,  273,   -1,   -1,  296,  297,   -1,
  279,   -1,  281,   -1,  303,   -1,   -1,  286,   -1,  288,
   -1,  290,   -1,  312,   -1,   -1,   -1,  296,  297,  263,
   -1,  265,   -1,  267,  303,  269,   -1,   -1,   -1,  273,
   -1,   -1,   -1,  312,   -1,  279,   -1,  281,   -1,  263,
   -1,  265,  286,  267,   -1,  269,  290,   -1,   -1,  273,
   -1,   -1,  296,  297,   -1,  279,   -1,  281,   -1,  303,
   -1,   -1,  286,   -1,   -1,   -1,  290,   -1,  312,   -1,
   -1,   -1,  296,  297,  263,   -1,  265,   -1,  267,  303,
  269,   -1,   -1,   -1,  273,   -1,   -1,   -1,  312,   -1,
  279,   -1,  281,   -1,   -1,   -1,   -1,  286,   -1,   -1,
   -1,  290,   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,
   -1,   -1,   -1,   -1,  303,   -1,  274,  275,  276,  277,
  278,   -1,   -1,  312,  282,  283,  284,  285,  286,  287,
  288,   -1,   -1,  291,  292,  274,  275,  276,  277,  278,
   -1,   -1,   -1,  282,  283,  284,  285,  286,  287,  288,
   -1,   -1,  291,  292,   -1,   -1,   -1,   -1,   -1,  317,
  270,  271,   -1,  302,  274,  275,  276,  277,  278,   -1,
   -1,   -1,  282,  283,  284,  285,  286,  287,  288,   -1,
   -1,  291,  292,  274,  275,  276,  277,  278,   -1,   -1,
   -1,  282,  283,  284,  285,  286,  287,  288,   -1,   -1,
  291,  292,  270,   -1,   -1,  296,  274,  275,  276,  277,
  278,   -1,   -1,   -1,  282,  283,  284,  285,  286,  287,
  288,   -1,   -1,  291,  292,  273,  274,  275,  276,  277,
  278,   -1,   -1,   -1,  282,  283,  284,  285,  286,  287,
  288,   -1,   -1,  291,  292,  274,  275,  276,  277,  278,
   -1,   -1,   -1,  282,  283,  284,  285,  286,  287,  288,
   -1,   -1,  291,  292,
  };

#line 881 "Parser.jay"

    public IProcess Parse(string file)
    {
      locationMap = new Dictionary<Object, int>();
      lexer = new Lexer(file);
      return (IProcess) yyparse(lexer);
    }

		public static bool HasLocation(Object obj)
		{
			return locationMap.ContainsKey(obj);
		}

    public static int GetLocation(Object obj)
    {
			Debug.Assert(locationMap.ContainsKey(obj), "no location for " + obj);
			return (int) locationMap[obj];
		}

		private static void SetLocation(Object obj, int location)
		{
		// Console.WriteLine("setting location {1} of {0}", obj, location);
			locationMap.Add(obj, location);
		}
}
#line default
namespace yydebug {
        using System;
	 internal interface yyDebug {
		 void push (int state, Object value);
		 void lex (int state, int token, string name, Object value);
		 void shift (int from, int to, int errorFlag);
		 void pop (int state);
		 void discard (int state, int token, string name, Object value);
		 void reduce (int from, int to, int rule, string text, int len);
		 void shift (int from, int to);
		 void accept (Object value);
		 void error (string message);
		 void reject ();
	 }
	 
	 class yyDebugSimple : yyDebug {
		 void println (string s){
			 Console.Error.WriteLine (s);
		 }
		 
		 public void push (int state, Object value) {
			 println ("push\tstate "+state+"\tvalue "+value);
		 }
		 
		 public void lex (int state, int token, string name, Object value) {
			 println("lex\tstate "+state+"\treading "+name+"\tvalue "+value);
		 }
		 
		 public void shift (int from, int to, int errorFlag) {
			 switch (errorFlag) {
			 default:				// normally
				 println("shift\tfrom state "+from+" to "+to);
				 break;
			 case 0: case 1: case 2:		// in error recovery
				 println("shift\tfrom state "+from+" to "+to
					     +"\t"+errorFlag+" left to recover");
				 break;
			 case 3:				// normally
				 println("shift\tfrom state "+from+" to "+to+"\ton error");
				 break;
			 }
		 }
		 
		 public void pop (int state) {
			 println("pop\tstate "+state+"\ton error");
		 }
		 
		 public void discard (int state, int token, string name, Object value) {
			 println("discard\tstate "+state+"\ttoken "+name+"\tvalue "+value);
		 }
		 
		 public void reduce (int from, int to, int rule, string text, int len) {
			 println("reduce\tstate "+from+"\tuncover "+to
				     +"\trule ("+rule+") "+text);
		 }
		 
		 public void shift (int from, int to) {
			 println("goto\tfrom state "+from+" to "+to);
		 }
		 
		 public void accept (Object value) {
			 println("accept\tvalue "+value);
		 }
		 
		 public void error (string message) {
			 println("error\t"+message);
		 }
		 
		 public void reject () {
			 println("reject");
		 }
		 
	 }
}
// %token constants
 class Token {
  public const int ERROR = 257;
  public const int TEXT = 258;
  public const int INT = 259;
  public const int FLOAT = 260;
  public const int ID = 261;
  public const int ARROW = 262;
  public const int DARROW = 263;
  public const int LBRACE = 264;
  public const int RBRACE = 265;
  public const int LPAREN = 266;
  public const int RPAREN = 267;
  public const int LBRACK = 268;
  public const int RBRACK = 269;
  public const int QMARK = 270;
  public const int EMARK = 271;
  public const int COLON = 272;
  public const int SEMICOLON = 273;
  public const int DOT = 274;
  public const int PLUS = 275;
  public const int MINUS = 276;
  public const int TIMES = 277;
  public const int DIVIDE = 278;
  public const int AMPERSAND = 279;
  public const int BACKSLASH = 280;
  public const int EQ = 281;
  public const int EQEQ = 282;
  public const int NE = 283;
  public const int LT = 284;
  public const int LE = 285;
  public const int GT = 286;
  public const int GE = 287;
  public const int COMMA = 288;
  public const int TILDE = 289;
  public const int OR = 290;
  public const int LAND = 291;
  public const int LOR = 292;
  public const int IN_CAP = 293;
  public const int OUT_CAP = 294;
  public const int INOUT_CAP = 295;
  public const int KW_IN = 296;
  public const int KW_AND = 297;
  public const int KW_BOOL = 298;
  public const int KW_INT = 299;
  public const int KW_FLOAT = 300;
  public const int KW_STRING = 301;
  public const int KW_WITH = 302;
  public const int KW_LOCATION = 303;
  public const int KW_WSDL = 304;
  public const int KW_NIL = 305;
  public const int KW_SCHEMA = 306;
  public const int KW_LET = 307;
  public const int KW_NEW = 308;
  public const int KW_JOIN = 309;
  public const int KW_SPAWN = 310;
  public const int KW_MATCH = 311;
  public const int KW_WHEN = 312;
  public const int KW_IMPORT = 313;
  public const int KW_FALSE = 314;
  public const int KW_TRUE = 315;
  public const int KW_IF = 316;
  public const int KW_THEN = 317;
  public const int KW_ELSE = 318;
  public const int Selectdec = 319;
  public const int UMINUS = 320;
  public const int yyErrorCode = 256;
 }
 namespace yyParser {
  using System;
  /** thrown for irrecoverable syntax errors and stack overflow.
    */
  internal class yyException : System.Exception {
    public yyException (string message) : base (message) {
    }
  }

  /** must be implemented by a scanner object to supply input to the parser.
    */
  internal interface yyInput {
    /** move on to next token.
        @return false if positioned beyond tokens.
        @throws IOException on input error.
      */
    bool advance (); // throws java.io.IOException;
    /** classifies current token.
        Should not be called if advance() returned false.
        @return current %token or single character.
      */
    int token ();
    /** associated with current token.
        Should not be called if advance() returned false.
        @return value for token().
      */
    Object value ();
  }
 }
} // close outermost namespace, that MUST HAVE BEEN opened in the prolog
